import { createClient } from 'redis';
import { RequestEvent, Handle } from '@sveltejs/kit';
import { Firestore } from 'firebase-admin/firestore';

declare const SECOND = 1;
declare const MINUTE: number;
declare const HOUR: number;
declare const DAY: number;
declare const WEEK: number;
/**
 * Refill represents a token bucket refill rate and size
 */
declare class Refill {
    readonly rate: number;
    readonly size: number;
    /**
     * Create a Refill instance
     *
     * @param rate Refill rate per second
     * @param size Total size (burst capacity)
     */
    constructor(rate: number, size?: number);
    static parse(value: string): Refill;
    private static unitToDenominator;
}

interface BucketResult {
    allowed: boolean;
    limit: number;
    remaining: number;
    reset: number;
    policy: string;
}
interface TokenBuckets {
    consume(key: string, refill: Refill, count?: number): Promise<BucketResult>;
}
declare abstract class TokenBucket {
    protected abstract consume(key: string, refill: Refill, count?: number): Promise<BucketResult>;
    protected apply(refill: Refill, count: number, allowed: boolean, tokens: number): {
        allowed: boolean;
        limit: number;
        remaining: number;
        reset: number;
        policy: string;
    };
}

declare class InMemoryTokenBucket extends TokenBucket {
    private readonly buckets;
    consume(key: string, refill: Refill, count?: number): Promise<{
        allowed: boolean;
        limit: number;
        remaining: number;
        reset: number;
        policy: string;
    }>;
}

type Redis$1 = ReturnType<typeof createClient>;
declare class RedisTokenBucket extends TokenBucket {
    private readonly redis;
    private readonly prefix;
    private readonly module;
    private readonly fname;
    constructor(redis: Redis$1, prefix?: string, module?: string, fname?: string);
    consume(key: string, refill: Refill, count?: number): Promise<{
        allowed: boolean;
        limit: number;
        remaining: number;
        reset: number;
        policy: string;
    }>;
    static create(redis: Redis$1, prefix?: string, module?: string, fname?: string): Promise<RedisTokenBucket>;
}

type CustomFn = (event: RequestEvent, key: string | null) => Promise<string | null>;
interface SearchParam {
    searchParam: string;
    httpHeader?: string;
    cookie?: string;
    custom?: CustomFn;
}
interface HttpHeader {
    searchParam?: string;
    httpHeader: string;
    cookie?: string;
    custom?: CustomFn;
}
interface Cookie {
    searchParam?: string;
    httpHeader?: string;
    cookie: string;
    custom?: CustomFn;
}
interface Custom {
    searchParam?: string;
    httpHeader?: string;
    cookie?: string;
    custom: CustomFn;
}
type Options = SearchParam | HttpHeader | Cookie | Custom;
declare class KeyExtractor {
    private readonly options;
    constructor(options: Options);
    extract(event: RequestEvent): Promise<string | null>;
}

interface KeyInfo {
    user: string;
    name: string;
    description: string;
    expires: Date | null;
    permissions: string[];
}

interface KeyStore {
    put(hash: string, info: KeyInfo): Promise<void>;
    get(hash: string): Promise<KeyInfo | null>;
    del(hash: string): Promise<void>;
    list(user: string): Promise<KeyInfo[]>;
}

declare class KeyManager {
    private readonly storage;
    private readonly key_length;
    constructor(storage: KeyStore, key_length?: number);
    private hashFromBytes;
    generate(info: KeyInfo): Promise<string>;
    validate(key: string | null): Promise<KeyInfo | null>;
    remove(key: string): Promise<void>;
    retrieve(user: string): Promise<KeyInfo[]>;
}

type Input = Parameters<Handle>[0];
declare class Handler {
    private readonly extractor;
    private readonly manager;
    private readonly bucket;
    constructor(extractor: KeyExtractor, manager: KeyManager, bucket: TokenBuckets);
    handle: (input: Input) => Promise<Response>;
}
declare class Api {
    private readonly event;
    private readonly bucket;
    readonly key: string | null;
    readonly info: KeyInfo | null;
    private _name;
    private _cost;
    private _anon;
    private _has;
    private _any;
    private _all;
    constructor(event: RequestEvent, bucket: TokenBuckets, key: string | null, info: KeyInfo | null);
    anonymous(): this;
    name(name: string): this;
    cost(cost: number): this;
    has(permission: string): this;
    any(permissions: string[]): this;
    all(permissions: string[]): this;
    approve(refill: Refill): Promise<BucketResult>;
}

declare class FirestoreKeyStore implements KeyStore {
    private readonly firestore;
    private readonly name;
    private readonly collection;
    constructor(firestore: Firestore, name?: string);
    put(hash: string, info: KeyInfo): Promise<void>;
    get(hash: string): Promise<KeyInfo | null>;
    del(hash: string): Promise<void>;
    list(user: string): Promise<KeyInfo[]>;
}

declare class InMemoryKeyStore implements KeyStore {
    private keys;
    put(hash: string, info: KeyInfo): Promise<void>;
    get(hash: string): Promise<KeyInfo | null>;
    del(hash: string): Promise<void>;
    list(user: string): Promise<KeyInfo[]>;
}

type Redis = ReturnType<typeof createClient>;
declare class RedisKeyStore implements KeyStore {
    private readonly redis;
    private readonly prefix;
    static create(redis: Redis, prefix?: string): Promise<RedisKeyStore>;
    private constructor();
    put(hash: string, info: KeyInfo): Promise<void>;
    get(hash: string): Promise<KeyInfo | null>;
    del(hash: string): Promise<void>;
    list(user: string): Promise<{
        user: string;
        name: string;
        description: string;
        permissions: string[];
        expires: Date | null;
    }[]>;
}

interface ApiLocals {
    api: Api;
}
declare global {
    namespace App {
        interface Locals extends ApiLocals {
        }
    }
}

export { Api, type ApiLocals, type BucketResult, type Cookie, type Custom, DAY, FirestoreKeyStore, HOUR, Handler, type HttpHeader, InMemoryKeyStore, InMemoryTokenBucket, KeyExtractor, type KeyInfo, KeyManager, type KeyStore, MINUTE, type Options, RedisKeyStore, RedisTokenBucket, Refill, SECOND, type SearchParam, TokenBucket, type TokenBuckets, WEEK };
